"""Function for averaging subject-level transition probabilities and producing visualizations."""

from typing import Optional, Union

import pandas as pd

from neurocaps.utils import PlotDefaults
from neurocaps.utils import _io as io_utils
from neurocaps.utils._helpers import resolve_kwargs
from neurocaps.utils._plot_utils import MatrixVisualizer, PlotFuncs


def transition_matrix(
    trans_dict: dict[str, pd.DataFrame],
    output_dir: Optional[str] = None,
    plot_output_format: str = "png",
    suffix_filename: Optional[str] = None,
    suffix_title: Optional[str] = None,
    save_plots: bool = True,
    save_df: bool = True,
    show_figs: bool = True,
    return_df: bool = True,
    **kwargs,
) -> Union[pd.DataFrame, None]:
    """
    Generate and Visualize the Averaged Transition Probabilities.

    Averages subject-level transition probabilities to produce a transition probability matrix. One
    matrix is generated per group.

    Parameters
    ----------
    trans_dict: :obj: `dict[str, pd.DataFrame]`
        A dictionary mapping groups to pandas DataFrame containing the transition probabilities for
        each subject. This assumes the output from ``CAP.calculate_metrics`` is being used,
        specifically ``metrics_output["transition_probability"]``.

    output_dir: :obj:`str` or :obj:`None`, default=None
        Directory to save plots (if ``save_plots`` is True) and transition probability matrices
        DataFrames (if ``save_df`` is True) to. The directory will be created if it does not exist.
        Plots and dataframes will not be saved if None.

    plot_output_format: :obj:`str`, default="png"
        The format to save plots in when ``output_dir`` is specified. Options are "png" or
        "pkl" (which can be further modified). Note that "pickle" is also accepted.

        .. versionchanged:: 0.33.0
            Replaces ``as_pickle`` and accepts a string value.

    suffix_filename: :obj:`str` or :obj:`None`, default=None
        Appended to the filename of each saved plot if ``output_dir`` is provided.

    suffix_title: :obj:`str` or :obj:`None`, default=None
        Appended to the title of each plot.

    save_plots: :obj:`bool`, default=True
        If True, plots are saves as png images. For this to be used, ``output_dir`` must be specified.

    save_df: :obj:`bool`, default=False,
        If True, saves the transition probability matrix contained in the DataFrames as csv files.
        For this to be used, ``output_dir`` must be specified.

    show_figs: :obj:`bool`, default=True
        Display figures.

    return_df: :obj:`bool`, default=False
        If True, returns a dictionary with a transition probability matrix for each group.

    **kwargs
        Additional keyword arguments for customizing plots.
        See :meth:`neurocaps.utils.PlotDefaults.transition_matrix` for all available options and
        their default values (See `PlotDefaults Documentation for transition_matrix\
        <https://neurocaps.readthedocs.io/en/stable/api/generated/neurocaps.utils.PlotDefaults.transition_matrix.html#neurocaps.utils.PlotDefaults.transition_matrix>`_)

    Returns
    -------
    dict[str, pd.DataFrame]
        An instance of a pandas DataFrame for each group if ``return_df`` is True.

    Note
    ----
    **Dataframe Representation**: Rows represent "from" and columns represent "to". For instance,
    the probability at ``df.loc["CAP-1", "CAP-2"]`` represents the averaged probability from
    transitioning from CAP-1 to CAP-2.

    +------------+---------+-------+-------+
    | From/To    |  CAP-1  | CAP-2 | CAP-3 |
    +============+=========+=======+=======+
    | CAP-1      |  0.40   | 0.35  | 0.25  |
    +------------+---------+-------+-------+
    | CAP-2      |  0.20   | 0.45  | 0.35  |
    +------------+---------+-------+-------+
    | CAP-3      |  0.35   | 0.18  |  0.47 |
    +------------+---------+-------+-------+
    """
    assert isinstance(
        trans_dict, dict
    ), "transition_dict must be in the form dict[str, pd.DataFrame]."

    io_utils.issue_file_warning("suffix_filename", suffix_filename, output_dir)

    # Create plot dictionary
    plot_dict = resolve_kwargs(PlotDefaults.transition_matrix(), **kwargs)

    trans_mat_dict = {}

    for group_name in trans_dict:
        df = trans_dict[group_name]
        # Get indices and averaged probabilities
        indices, averaged_probabilities = df.iloc[:, 3:].mean().index, df.iloc[:, 3:].mean().values
        # Get the maximum CAP
        max_cap = str(max(float(i) for i in indices)).split(".")[0]
        cap_names = [f"CAP-{num}" for num in range(1, int(max_cap) + 1)]
        trans_mat = pd.DataFrame(index=cap_names, columns=cap_names, dtype="float64")
        # Add name to index
        trans_mat.index.name = "From/To"

        # Create matrix
        for location, name in enumerate(indices):
            trans_mat.loc[f"CAP-{name.split('.')[0]}", f"CAP-{name.split('.')[1]}"] = (
                averaged_probabilities[location]
            )

        display = MatrixVisualizer.create_display(
            trans_mat, plot_dict, suffix_title, group_name, "trans"
        )

        trans_mat_dict[group_name] = trans_mat

        # Save figure & dataframe
        if output_dir:
            MatrixVisualizer.save_contents(
                display=display,
                plot_dict=plot_dict,
                output_dir=output_dir,
                plot_output_format=plot_output_format,
                suffix_filename=suffix_filename,
                group_name=group_name,
                curr_dict=trans_mat_dict,
                save_plots=save_plots,
                save_df=save_df,
                call="transition_matrix",
            )

        PlotFuncs.show(show_figs)

    if return_df:
        return trans_mat_dict



"""Internal helper functions."""

from typing import Any

from ._logging import setup_logger

LG = setup_logger(__name__)


def resolve_kwargs(defaults: dict[str, Any], **kwargs) -> dict[str, Any]:
    plot_dict = defaults.copy()
    plot_dict.update({k: v for k, v in kwargs.items() if k in plot_dict})

    if kwargs:
        invalid_kwargs = {k: v for k, v in kwargs.items() if k not in plot_dict}
        if invalid_kwargs:
            LG.info(
                f"The following invalid kwargs arguments used and will be ignored: {invalid_kwargs}"
            )

    return plot_dict


def list_to_str(str_list: list[str]) -> None:
    """Converts a list containing strings to a string."""
    return ", ".join(["'{a}'".format(a=x) for x in str_list])



"""Internal functions for extracting timeseries with or without joblib."""

import copy, json, inspect, math, os, re
from dataclasses import dataclass, field
from functools import cached_property
from typing import Union

import numpy as np, pandas as pd
from nilearn.maskers import NiftiLabelsMasker
from nilearn.image import index_img, load_img
from scipy.interpolate import CubicSpline

from neurocaps.typing import ParcelApproach
from neurocaps.utils._logging import setup_logger

# Logger initialization to check if any user-defined loggers where created prior to package import.
# No variable assignment needed.
setup_logger(__name__)


@dataclass
class RunData:
    """
    A data class that operates as a container to be passed to various helper functions.
    Contains run-specific information.
    """

    # Add defaults so no field are required when using this class for testing
    parcel_approach: ParcelApproach = field(default_factory=dict)
    signal_clean_info: dict = field(default_factory=dict)
    task_info: dict = field(default_factory=dict)
    tr: Union[int, None] = None
    verbose: bool = False
    # Run-specific attributes
    files: dict[str, str] = field(default_factory=dict)
    head: Union[str, None] = None
    dummy_vols: Union[int, None] = None
    censored_frames: list[int] = field(default_factory=list)
    sample_mask: Union[np.typing.NDArray[np.bool_], None] = None
    censored_ends: list[int] = field(default_factory=list)
    fd_array_len: Union[int, None] = None
    # Event condition scan indices and qc information for condition
    scans: list[int] = field(default_factory=list)
    n_total_scans: Union[int, None] = None
    n_censored_scans: Union[int, None] = None
    n_interpolated_scans: Union[int, None] = None
    # Avoid timeseries extraction
    skip_run: bool = False
    # Stats for qc
    mean_fd: Union[float, None] = None
    std_fd: Union[float, None] = None
    high_motion_len_mean: Union[float, None] = None
    high_motion_len_std: Union[float, None] = None

    @property
    def session(self) -> Union[int, str, None]:
        return self.task_info["session"]

    @property
    def task(self) -> str:
        return self.task_info["task"]

    @property
    def condition(self) -> Union[str, None]:
        return self.task_info["condition"]

    @property
    def tr_shift(self) -> int:
        return self.task_info["condition_tr_shift"]

    @property
    def slice_ref(self) -> float:
        return self.task_info["slice_time_ref"]

    @property
    def use_confounds(self) -> bool:
        return self.signal_clean_info["use_confounds"]

    @property
    def confound_names(self) -> Union[list[str], None]:
        return self.signal_clean_info["confound_names"]

    @property
    def n_acompcor_separate(self) -> Union[int, None]:
        return self.signal_clean_info["n_acompcor_separate"]

    @property
    def fd_thresh(self) -> Union[float, None]:
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return self.signal_clean_info["fd_threshold"]["threshold"]
        else:
            return self.signal_clean_info["fd_threshold"]

    @property
    def out_percent(self) -> Union[float, None]:
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return self.signal_clean_info["fd_threshold"].get("outlier_percentage")

    @property
    def censored_vals(self) -> tuple[int, int]:
        return (
            (self.n_censored_scans, self.n_total_scans)
            if self.condition
            else (len(self.censored_frames), self.fd_array_len)
        )

    @property
    def n_before(self) -> Union[int, None]:
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return self.signal_clean_info["fd_threshold"].get("n_before")

    @property
    def n_after(self) -> Union[int, None]:
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return self.signal_clean_info["fd_threshold"].get("n_after")

    @property
    def pass_mask_to_nilearn(self) -> Union[bool, None]:
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return (
                True
                if self.signal_clean_info["fd_threshold"].get("use_sample_mask") is True
                else False
            )
        else:
            return False

    @property
    def interpolate(self) -> Union[bool, None]:
        if isinstance(self.signal_clean_info["fd_threshold"], dict):
            return self.signal_clean_info["fd_threshold"].get("interpolate")

    @cached_property
    def confound_df(self) -> Union[pd.DataFrame, None]:
        if self.files["confound"]:
            return pd.read_csv(self.files["confound"], sep="\t")
        else:
            return None

    @property
    def maps(self) -> str:
        return self.parcel_approach[list(self.parcel_approach)[0]]["maps"]


def process_subject_runs(
    subj_id,
    prepped_files,
    run_list,
    parcel_approach,
    signal_clean_info,
    task_info,
    tr,
    verbose,
    parallel_log_config=None,
):
    """
    Entry point for extracting timeseries data and getting quality control information on a per run
    basis. Before the actual extraction is conducted, the framewise displacement censoring
    information (if requested) and the condition event information (if requested) are obtained to
    determine if a run should be skipped. This is done to prevent extraction of timeseries where all
    frames are censored, too many frames are censored (based on "outlier_percentage"), or no
    condition can be extracted.

    Note
    ----
    A default id of "run-0" is given to data where there is no run ID.
    """
    # Logger inside function to give logger to each child process if parallel processing is done
    LG = setup_logger(__name__, top_level=False, parallel_log_config=parallel_log_config)

    # Initialize subject dictionary and quality control dictionary
    subject_timeseries = {subj_id: {}}
    qc = {subj_id: {}}

    for run in run_list:
        # Initialize class; placing inside run loops allows re-initialization of defaults
        data = RunData(parcel_approach, signal_clean_info, task_info, tr, verbose)

        run_id, data.files, data.head = get_subject_data(subj_id, run, prepped_files, data, LG)

        # Get dummy volumes
        data.dummy_vols = get_dummy(data, LG)

        # Assess framewise displacement
        col_name = "framewise_displacement"
        if data.fd_thresh and data.files["confound"] and col_name in data.confound_df.columns:
            data.censored_frames, data.fd_array_len, data.skip_run = compute_censored_frames(
                data, LG
            )

            # All frames exceed threshold
            if data.skip_run:
                continue

            data.sample_mask = create_sample_mask(data)

            if not data.condition:
                data.mean_fd, data.std_fd, data.high_motion_len_mean, data.high_motion_len_std = (
                    get_motion_stats(data)
                )

        elif data.fd_thresh and data.files["confound"] and col_name not in data.confound_df.columns:
            LG.warning(
                f"{data.head}" + "`fd_threshold` specified but 'framewise_displacement' column not "
                "found in the confound tsv file. Removal of volumes after nuisance regression will "
                "not be but timeseries extraction will continue."
            )

        # Determines indxs of ends to be censored if interpolation of censored volumes are requested
        if data.interpolate:
            data.censored_ends = get_contiguous_censored_ends(data.sample_mask)

        # Get condition windows
        if data.files["event"]:
            condition_df, data.skip_run = get_condition_df(data, LG)

            # Condition df is empty
            if data.skip_run:
                continue

            data.scans, data.n_total_scans = get_condition_indices(data, condition_df)

            if data.censored_frames or data.fd_thresh:
                # Use fd array length to determine if indx is out of bounds
                data.scans = validate_scan_bounds(data, data.fd_array_len, LG)
                data.mean_fd, data.std_fd, data.high_motion_len_mean, data.high_motion_len_std = (
                    get_motion_stats(data)
                )

                if data.censored_frames:
                    # Removing censored or non-interpolated scan indxs
                    data.scans, data.n_censored_scans, data.n_interpolated_scans = (
                        filter_censored_scan_indices(data)
                    )
                else:
                    data.n_censored_scans, data.n_interpolated_scans = 0, 0

            if not data.scans:
                LG.warning(
                    f"{data.head}"
                    + f"[CONDITION: {data.condition}] Timeseries Extraction Skipped: Timeseries "
                    "will be empty when filtered to only include volumes from the requested "
                    "condition. Possibly due to the TRs corresponding to the condition being "
                    "removed by `dummy_scans` or filtered due to exceeding threshold for "
                    "`fd_threshold`."
                )
                continue

        # Compute number of volumes exceeding outlier percentage
        if data.censored_frames and data.out_percent is not None:
            vols_exceed_percent, data.skip_run = flag_run(data)

            if data.skip_run:
                log_high_motion_percentage(data, LG, vols_exceed_percent)
                continue

        timeseries, data.skip_run = perform_extraction(data, LG)

        if data.skip_run:
            continue

        if timeseries.shape[0] == 0:
            LG.warning(
                f"{data.head}" + f"Timeseries is empty and will not be appended to the "
                "`subject_timeseries` dictionary."
            )
        else:
            subject_timeseries[subj_id].update({run_id: timeseries})

            qc[subj_id].update({run_id: report_qc(data)})

    if not subject_timeseries[subj_id]:
        LG.warning(
            f"{data.head.split(' | RUN:')[0]}] Timeseries Extraction Skipped: No runs were "
            "extracted."
        )
        subject_timeseries, qc = None, None

    if qc is not None and not qc[subj_id]:
        qc = None

    return subject_timeseries, qc


def get_subject_data(subj_id, run, prepped_files, data, LG):
    """Gets subject-related data."""
    run_id = "run-0" if run is None else run

    files = {
        "nifti": grab_file(run, prepped_files["niftis"]),
        "confound": grab_file(run, prepped_files["confounds"]),
        "confound_meta": grab_file(
            run, prepped_files["confound_metas"] if prepped_files.get("confound_metas") else None
        ),
        "event": grab_file(run, prepped_files["events"]),
    }

    # Base message containing subject header information for logging
    head = subject_header(data, run_id.split("-")[-1], subj_id, files["nifti"])

    if data.verbose:
        LG.info(
            f"{head}" + "Preparing for Timeseries Extraction using "
            f"[FILE: {os.path.basename(files['nifti'])}]."
        )

    return run_id, files, head


def grab_file(run, files):
    """
    Grabs a single from a list of files. If ``files`` is not None, then the file corresponding to
    the run ID is returned. If ``run`` is None, it's assumed that their is only a single run in
    ``files``.
    """
    if not files:
        return None

    if run:
        return [file for file in files if f"{run}_" in os.path.basename(file)][0]
    else:
        return files[0]


def subject_header(data, run_id, subj_id, nifti):
    """
    Creates the subject header to use in verbose logging, indicating the subject, session, task,
    and run.
    """
    if data.session:
        sess_id = data.session
    else:
        base_filename = os.path.basename(nifti)
        sess = (
            re.search("ses-(\\S+?)[-_]", base_filename)[0][:-1] if "ses-" in base_filename else None
        )
        sess_id = sess.split("-")[-1] if sess else None

    sub_head = f"[SUBJECT: {subj_id} | SESSION: {sess_id} | TASK: {data.task} | RUN: {run_id}]"

    return f"{sub_head} "


def get_dummy(data, LG):
    """Gets the number of dummy scans to remove."""
    info = data.signal_clean_info["dummy_scans"]

    use_auto = info == "auto" or (isinstance(info, dict) and info.get("auto"))
    if use_auto:
        n, flag = int(data.confound_df.columns.str.startswith("non_steady_state").sum()), "auto"
    else:
        n, flag = data.signal_clean_info["dummy_scans"], None

    if use_auto and isinstance(info, dict):
        if info.get("min") and n < info["min"]:
            n, flag = info["min"], "min"
        if info.get("max") and n > info["max"]:
            n, flag = info["max"], "max"

    if all([use_auto, flag, data.verbose]):
        if flag == "auto":
            if n:
                LG.info(
                    f"{data.head}" + "Number of dummy scans to be removed based on "
                    f"'non_steady_state_outlier_XX' columns: {n}."
                )
            else:
                LG.info(
                    f"{data.head}" + "No 'non_steady_state_outlier_XX' columns were found so 0 "
                    "dummy scans will be removed."
                )
        else:
            LG.info(f"{data.head}" + f"Default dummy scans set by '{flag}' will be used: {n}.")

    return n


def compute_censored_frames(data, LG):
    """Determines the corresponding indices that exceed a specific fd threshold."""
    skip_run = False

    # Get censored frames and length of fd array (minus dummy volumes)
    censored_frames, fd_array_len = basic_censor(data)

    if censored_frames and (data.n_before or data.n_after):
        censored_frames = extended_censor(censored_frames, fd_array_len, data)

    if len(censored_frames) == fd_array_len:
        LG.warning(
            f"{data.head}" + "Timeseries Extraction Skipped: Timeseries will be empty due to "
            f"all volumes exceeding a framewise displacement of {data.fd_thresh}."
        )
        skip_run = True

    return censored_frames, fd_array_len, skip_run


def basic_censor(data):
    """
    Finds the indices that exceed a certain framewise displacement threshold after dummy volumes are
    removed.
    """
    fd_array = get_fd_array(data)
    censor_volumes = sorted(list(np.where(fd_array > data.fd_thresh)[0]))

    return censor_volumes, fd_array.shape[0]


def get_fd_array(data):
    """Retrieves the fd array with dummy volumes removed."""
    # First volume may be nan
    fd_array = data.confound_df["framewise_displacement"].fillna(0).values

    if data.dummy_vols:
        fd_array = fd_array[data.dummy_vols :]

    return fd_array


def extended_censor(censored_frames, fd_array_len, data):
    """
    Iterates through each element in ``censored_frames`` (representing a high motion index) and
    computes the indices of the frames before ``i`` to also censor ("n_before") as well as the
    indices after ``i`` to censor.
    """
    ext_arr = []

    for i in censored_frames:
        if data.n_before:
            ext_arr.extend(range(i - data.n_before, i))
        if data.n_after:
            ext_arr.extend(range(i + 1, i + data.n_after + 1))

    filtered_ext_arr = [x for x in ext_arr if x >= 0 and x < fd_array_len]

    return sorted(list(set(censored_frames + filtered_ext_arr)))


def create_sample_mask(data):
    """
    Creates a boolean using ``data.fd_array_len``, which is the length of the ``fd_array``.
    Assumes dummy volumes are removed. In this mask, False (0) are the censored volumes and True
    (1) are the retained volumes.
    """
    sample_mask = np.ones(data.fd_array_len, dtype="bool")

    if data.censored_frames:
        sample_mask[np.array(data.censored_frames)] = False

    return sample_mask


def get_contiguous_censored_ends(sample_mask):
    """
    Assumes dummy scans have already been filtered out of ``sample_mask`` by the time``sample_mask``
    is created. Determines if the contiguous edges of the full timeseries have been flagged.
    """
    if np.all(sample_mask):
        return []

    contiguous_indices = get_contiguous_segments(sample_mask)
    # Check if first and last index in sample mask is 0
    start_indices = contiguous_indices[0].tolist() if sample_mask[0] == 0 else []
    end_indices = contiguous_indices[-1].tolist() if sample_mask[-1] == 0 else []

    return start_indices + end_indices


def get_contiguous_segments(sample_mask, splice="indices"):
    """
    Get contiguous segments of high motion and low motion data for the sample mask and return as
    indices or booleans.

    Example
    -------

    >>> import numpy as np
    >>> arr = np.array([0, 0, 1, 1, 0, 0, 1, 1, 0, 0]) # sample mask
    >>> diff_arr = np.diff(arr) # [0, 1, 0, -1, 0, 1, 0, -1, 0]
    >>> indxs = np.where(diff_arr != 0)[0] + 1 # [1, 3 ,5 , 7] -> [2, 4, 6, 8]
    >>> # arr = np.arange(sample_mask.shape[0]) if splitting on indices
    >>> segments = np.split(arr, indxs) # Groups of binaries or indices
    >>> print(segments)
        [array([0, 0]), array([1, 1]), array([0, 0]), array([1, 1]), array([0, 0])]
    """
    split_indices = np.where(np.diff(sample_mask, n=1) != 0)[0] + 1
    arr = np.arange(sample_mask.shape[0]) if splice == "indices" else sample_mask
    segments = np.split(arr, split_indices)

    return segments


def get_motion_stats(data):
    """
    Get mean and standard deviation of FD prior to scrubbing, as well as the mean and standard
    deviation of the average flagged frames, regardless if frames are to be interpolated or
    scrubbed. Bessel's correction not used for standard deviation.

    Note
    ----
    For computational simplicity, scan indices are treated as the true, continuous length of the
    timeseries so gaps between event windows are not considered.
    """
    sample_mask = data.sample_mask[data.scans] if data.condition else data.sample_mask

    fd_array = get_fd_array(data)
    fd_array = fd_array[data.scans] if data.condition else fd_array
    mean_fd, std_fd = np.mean(fd_array), np.std(fd_array, ddof=0)

    if np.all(sample_mask):
        return mean_fd, std_fd, 0, 0

    high_motion_segments = get_contiguous_segments(sample_mask, splice="sample_mask")
    high_motion_segment_counts = [
        len(segment) for segment in high_motion_segments if segment[0] == 0
    ]

    return (
        mean_fd,
        std_fd,
        np.mean(high_motion_segment_counts),
        np.std(high_motion_segment_counts, ddof=0),
    )


def get_condition_df(data, LG):
    """
    Obtains a dataframe only containing information for a specific condition specified in
    "trail type".
    """
    skip_run = False

    event_df = pd.read_csv(data.files["event"], sep="\t")
    condition_df = event_df[event_df["trial_type"] == data.condition]

    if condition_df.empty:
        LG.warning(
            f"{data.head}" + f"[CONDITION: {data.condition}] Timeseries Extraction Skipped: The "
            "requested condition does not exist in the 'trial_type' column of the event file."
        )
        skip_run = True

    return condition_df, skip_run


def get_condition_indices(data, condition_df):
    """
    Converts condition event timing to TR units to be extracted from the timeseries. Adjusts these
    TRs based on slice timing (if specified, to adjust the onset time back) and a tr shift
    (if specified, to adjust for hemodynamic lag). Also makes an adjustment for dummy volumes
    (if specified).
    """
    scans = []

    for i in condition_df.index:
        adjusted_onset = condition_df.loc[i, "onset"] - data.slice_ref * data.tr
        onset_scan = math.floor(adjusted_onset / data.tr) + data.tr_shift
        end_scan = (
            math.ceil((adjusted_onset + condition_df.loc[i, "duration"]) / data.tr) + data.tr_shift
        )

        # Avoid accidental negative indexing
        onset_scan = max([0, onset_scan])
        end_scan = max([0, end_scan])
        scans.extend(range(onset_scan, end_scan))

    scans = sorted(list(set(scans)))

    if data.dummy_vols:
        scans = [scan - data.dummy_vols for scan in scans if scan not in range(data.dummy_vols)]

    return scans, len(scans)


def filter_censored_scan_indices(data):
    """
    Removes condition indices, stored in ``data.scans``, that exceed the length of the ``fd_array``,
    which corresponds to the timeseries length. Also removed any indices that are censored and
    provides the number of indices to be scrubbed or interpolated.
    """
    scans = set(data.scans)

    all_censored_indxs = scans.intersection(data.censored_frames)
    # Determine the number of frames that will be interpolated for qc report
    n_interpolated_scans = (
        len(all_censored_indxs.difference(data.censored_ends)) if data.interpolate else 0
    )
    # Remove all censored scans if no interpolation else only remove censored (contiguous) ends
    scrubbing_approach = data.censored_frames if not data.interpolate else data.censored_ends
    final_censored_indxs = scans.intersection(scrubbing_approach)
    scans = scans.difference(final_censored_indxs)

    return sorted(list(scans)), len(all_censored_indxs), n_interpolated_scans


def validate_scan_bounds(data, arr_len, LG=None):
    """
    Checks if indices in ``data.scans`` (which are the indices for the event condition) are within
    the timeseries boundaries.
    """
    if max(data.scans) <= arr_len - 1:
        return data.scans

    # Warn if tr_shift was not done, the warning is specifically for instances related to
    # misalignment such as potentially incorrect onsets and durations
    if not data.tr_shift:
        LG.warning(
            f"{data.head}"
            + f"[CONDITION: {data.condition}] Max scan index exceeds timeseries max index. "
            f"Max condition index is {max(data.scans)}, while max timeseries index is "
            f"{arr_len - 1}. Timing may be misaligned or specified repetition time incorrect. If "
            "intentional, ignore warning. Only indices for condition within the timeseries range "
            "will be extracted."
        )

    scans = set(data.scans).intersection(range(arr_len))

    return sorted(list(scans))


def flag_run(data):
    """
    Determines if a run is flagged due to exceeding a certain threshold percentage specified by
    "outlier_percentage".
    """
    n_censor, n = data.censored_vals
    vols_exceed_percent = n_censor / n
    skip_run = vols_exceed_percent > data.out_percent

    return vols_exceed_percent * 100, skip_run


def log_high_motion_percentage(data, LG, vols_exceed_percent):
    """
    Logs the percentage of volumes exceeding a specified framewise displacement.
    """
    msg = f"Percentage of volumes exceeding the threshold limit is {vols_exceed_percent}%"

    if data.condition:
        msg += f" for [CONDITION: {data.condition}]"

    LG.warning(
        f"{data.head}" + f"Timeseries Extraction Skipped: Run flagged due to more than "
        f"{data.out_percent * 100}% of the volumes exceeding the framewise "
        f"displacement threshold of {data.fd_thresh}. {msg}."
    )


def perform_extraction(data, LG):
    """
    Pipeline to get the confounds passed to ``NiftiLabelsMasker`` for nuisance regression, extract
    the timeseries, and use ``data.scans`` to extract the condition if ``data.condition`` is not
    False.
    """
    # Extract confound information of interest and ensure confound file does not contain NAs
    confounds = process_confounds(data, LG)

    if data.use_confounds and (data.confound_names and confounds is None):
        LG.warning(
            f"{data.head}"
            + "None of the requested confounds were found. Nuisance regression will not be done "
            "but timeseries extraction will continue."
        )

    # Ensure number of confounds not greater than timeseries shape
    if skip_run := check_dimensions(data, confounds, LG):
        return None, skip_run

    masker = NiftiLabelsMasker(
        labels_img=data.maps,
        resampling_target="data",
        strategy="mean",
        standardize=False,
        t_r=data.tr,
        **data.signal_clean_info["masker_init"],
        **clean_param(),
    )

    # Load and discard volumes if needed
    nifti_img = load_img(data.files["nifti"], dtype=data.signal_clean_info["dtype"])

    if data.dummy_vols:
        nifti_img = index_img(nifti_img, slice(data.dummy_vols, None))
        if confounds is not None:
            confounds.drop(range(data.dummy_vols), axis=0, inplace=True)

    # Extract timeseries; Censor mask used only when `pass_mask_to_nilearn` is True
    timeseries = masker.fit_transform(
        nifti_img,
        confounds=confounds,
        sample_mask=(
            data.sample_mask if (data.pass_mask_to_nilearn and data.censored_frames) else None
        ),
    )

    del nifti_img

    # Process timeseries if censoring done;
    timeseries = process_timeseries(timeseries, data) if data.censored_frames else timeseries

    if data.condition:
        # Ensure condition indices don't exceed timeseries due to tr shift
        data.scans = validate_scan_bounds(data, timeseries.shape[0], LG)

        if data.verbose:
            LG.info(
                data.head
                + f"Nuisance regression completed; extracting [CONDITION: {data.condition}]."
            )

        # Extract condition
        timeseries = timeseries[data.scans, :]

    return (
        standardize_rois(timeseries) if data.signal_clean_info.get("standardize") else timeseries
    ), skip_run


def process_confounds(data, LG):
    """
    If there are valid names in ``data.signal_clean_info["confound_names"]`` or separate aCompCor
    components are requested, then this function obtains a filtered version of the confound
    dataframe containing columns that will be used for nuisance regression.
    """
    if not data.use_confounds or (data.use_confounds and data.confound_df.empty):
        return None

    confound_names = (
        copy.deepcopy(data.signal_clean_info["confound_names"]) if data.confound_names else []
    )

    # Extract first "n" numbers of specified WM and CSF components and extend confound_names list
    components_list = []
    if data.files["confound_meta"]:
        components_list = get_separate_acompcor(data)

    if components_list:
        confound_names.extend(components_list)

    confounds = extract_valid_confounds(data, confound_names, LG) if confound_names else None

    return confounds


def get_separate_acompcor(data):
    """
    Extract the first "n" numbers of specified WM and CSF components form the confounds metadata
    json file. Names are used to extract later to extract columns from the confounds dataframe.
    """
    confound_metadata_file = data.files["confound_meta"]
    components_list = []

    with open(confound_metadata_file, "r") as confounds_json:
        confound_metadata = json.load(confounds_json)

    acompcors = sorted([acompcor for acompcor in confound_metadata if "a_comp_cor" in acompcor])
    CSF = [CSF for CSF in acompcors if confound_metadata[CSF]["Mask"] == "CSF"][
        : data.n_acompcor_separate
    ]
    WM = [WM for WM in acompcors if confound_metadata[WM]["Mask"] == "WM"][
        : data.n_acompcor_separate
    ]
    components_list.extend(CSF + WM)

    return components_list


def extract_valid_confounds(data, confound_names, LG):
    """
    Extracts the confounds, specified in ``confound_names`` from the dataframe.
    """
    valid_confounds = []
    invalid_confounds = []
    col_names = data.confound_df.columns

    for confound_name in confound_names:
        if "*" in confound_name:
            prefix = confound_name.split("*")[0]
            confounds_list = list(col_names[col_names.str.startswith(prefix)])
        else:
            confounds_list = [confound_name] if confound_name in col_names else None

        if confounds_list:
            valid_confounds.extend(confounds_list)
        else:
            invalid_confounds.extend([confound_name])

    # First index of some variables is NaN
    confounds = data.confound_df[valid_confounds].fillna(0)
    confounds = None if confounds.empty else confounds

    if data.verbose and confounds is not None:
        if invalid_confounds:
            LG.warning(
                f"{data.head}"
                + f"The following confounds were not found: {', '.join(invalid_confounds)}."
            )

        LG.info(
            f"{data.head}" + "The following confounds will be used for nuisance regression: "
            f"{', '.join(list(confounds.columns))}."
        )

    return confounds


def check_dimensions(data, confounds, LG):
    """
    Prior to nuisance regression, Check if there are more regressors than volumes/frames
    (underdetermined) or if they are equal (exactly determined). Avoid case of zeroed out timeseries
    due to meeting one of these conditions. Does not consider linear dependence; however, meeting
    these criteria should be relatively rare.
    """
    if confounds is None:
        return False

    # Use the length of the fd_array for signal shape; Already accounts for dummy volume
    if data.pass_mask_to_nilearn and data.censored_frames:
        # Nilearn censors before removing signal variance associated with confounds
        signal_shape = data.fd_array_len - len(data.censored_frames)
    else:
        signal_shape = confounds.shape[0] - (0 if not data.dummy_vols else data.dummy_vols)

    if confounds.shape[1] >= signal_shape:
        LG.warning(
            f"{data.head}" + "Timeseries Extraction Skipped: The number of confound regressors "
            f"(n={confounds.shape[1]}) is equal to or greater than the timeseries shape "
            f"(n={signal_shape})."
        )
        return True
    else:
        return False


def clean_param():
    """
    Account for deprecation and eventual replacement of kwargs in nilearn versions greater than
    0.11.1 for ``clean_args``. Ensures that extrapolation is set to False when a sample mask is
    passed to nilearn since interpolation is done to deal with gaps when the Butterworth filter is
    used.
    """

    if "clean_args" in inspect.signature(NiftiLabelsMasker).parameters.keys():
        kwarg = {"clean_args": {"extrapolate": False}}
    else:
        kwarg = {"clean__extrapolate": False}

    return kwarg


def process_timeseries(timeseries, data):
    """
    Processes the timeseries by performing padding, interpolation, or removal of censored frames.
    """
    timeseries = pad_timeseries(timeseries, data)

    if data.interpolate:
        timeseries = interpolate_censored_frames(timeseries, data)
    else:
        timeseries = remove_censored_frames(timeseries, data.condition, data.censored_frames)

    return timeseries


def pad_timeseries(timeseries, data):
    """
    Temporarily pads the timeseries if a sample mask was passed to ``NiftiLabelsMasker``. Done to
    extract the correct indices corresponding to the event condition in the timeseries or to
    interpolate the correct frames.
    """
    # Early return if the sample mask was not passed to nilearn
    if not data.pass_mask_to_nilearn:
        return timeseries

    padded_timeseries = np.zeros((data.fd_array_len, timeseries.shape[1]))
    padded_timeseries[data.sample_mask, :] = timeseries

    return padded_timeseries


def interpolate_censored_frames(timeseries, data):
    """
    Replaces censored frames in the timeseries using cubic spline interpolation using scipy's
    ``CubicSpline``. Removes any censored frames that are not neighbored by non-censored data on
    its left or right.

    References `nilearn's _interpolate_volumes <https://github.com/nilearn/nilearn/blob/f74c4c5c0/nilearn/signal.py#L894>`_

    .. important:: Modifies ``timeseries`` in place.
    """
    # Get frame times
    frame_times = np.arange(data.fd_array_len) * data.tr
    # Create interpolated timeseries; retain only good frame times and timeseries data
    cubic_spline = CubicSpline(
        frame_times[data.sample_mask], timeseries[data.sample_mask, :], extrapolate=False
    )
    # Only replace the high motion volumes with the interpolated data
    timeseries[~data.sample_mask, :] = cubic_spline(frame_times)[~data.sample_mask, :]

    # Remove ends if not condition, condition already removes these volumes in the data.scans list
    return remove_censored_frames(timeseries, data.condition, data.censored_ends)


def remove_censored_frames(timeseries, condition, removed_frames):
    """
    Removes censored frames from the timeseries if no condition is specified since data.scans
    already ignores censored frames.
    """
    return timeseries if condition else np.delete(timeseries, removed_frames, axis=0)


def standardize_rois(timeseries, return_parameters=False):
    """
    Standardizes ROIs of timeseries if censoring or condition extraction occurs. Uses Bessel's
    correction (n-1). Standard deviations below ``np.finfo(std.dtype).eps`` are replaced with 1 for
    numerical stability.
    """
    mean = np.mean(timeseries, axis=0)
    std = np.std(timeseries, axis=0, ddof=1)
    std[std < np.finfo(std.dtype).eps] = 1.0
    timeseries = (timeseries - mean) / std

    return timeseries if not return_parameters else (timeseries, mean, std)


def report_qc(data):
    """
    Determines the number of frames that were scrubbed or interpolated and includes the stats of
    flagged frames.
    """
    report_val = lambda val: val if val is not None else math.nan
    qc = {
        "mean_fd": report_val(data.mean_fd),
        "std_fd": report_val(data.std_fd),
        "frames_scrubbed": math.nan,
        "frames_interpolated": math.nan,
        "mean_high_motion_length": report_val(data.high_motion_len_mean),
        "std_high_motion_length": report_val(data.high_motion_len_std),
        "n_dummy_scans": data.dummy_vols if data.signal_clean_info["dummy_scans"] else math.nan,
    }

    if not data.fd_thresh:
        return qc

    # QC report specific to the frames censored in the condition instead of overall censored frames
    if not data.condition:
        n_scrubbed_frames = len(data.censored_frames)
        n_interpolated_frames = (
            n_scrubbed_frames - len(data.censored_ends) if data.interpolate else 0
        )
        n_scrubbed_frames -= n_interpolated_frames
    else:
        n_scrubbed_frames = data.n_censored_scans - data.n_interpolated_scans
        n_interpolated_frames = data.n_interpolated_scans

    qc["frames_scrubbed"] = n_scrubbed_frames
    qc["frames_interpolated"] = n_interpolated_frames

    return qc



"""A class which is responsible for accessing attributes in ``TimeSeriesExtractor``."""

import copy, sys
from typing import Union

import numpy as np

from neurocaps.typing import ParcelConfig, ParcelApproach, SubjectTimeseries
from neurocaps.utils import _io as io_utils
from neurocaps.utils._parcellation_validation import check_parcel_approach


class TimeseriesExtractorGetter:
    def __init__(self):
        pass

    #### Exists upon initialization of TimeseriesExtractor
    @property
    def space(self) -> str:
        """
        The standard template space that the preprocessed BOLD data is registered to. This property
        is also settable.
        """
        return self._space

    @space.setter
    def space(self, new_space: str) -> None:
        if not isinstance(new_space, str):
            raise TypeError("`space` must be a string.")
        self._space = new_space

    @property
    def parcel_approach(self) -> ParcelApproach:
        """
        Parcellation information with "maps" (path to parcellation file), "nodes" (labels), and
        "regions" (anatomical regions or networks). Returns a deep copy.
        """
        return copy.deepcopy(self._parcel_approach)

    @parcel_approach.setter
    def parcel_approach(self, parcel_dict: Union[ParcelConfig, ParcelApproach, str]) -> None:
        self._parcel_approach = check_parcel_approach(parcel_approach=parcel_dict, call="setter")

    @property
    def signal_clean_info(self) -> Union[dict[str, Union[bool, int, float, str]], None]:
        """
        Dictionary containing signal cleaning parameters. Returns a deep copy.
        """
        return copy.deepcopy(self._signal_clean_info)

    ### Does not exists upon initialization of Timeseries Extractor

    # Exist when TimeSeriesExtractor.get_bold() used
    @property
    def task_info(self) -> Union[dict[str, Union[str, int]], None]:
        """
        Dictionary containing all task-related information such. Defined after running
        ``self.get_bold()``.
        """
        return getattr(self, "_task_info", None)

    # Gets initialized and populated in TimeSeriesExtractor.get_bold(),
    @property
    def subject_ids(self) -> Union[list[str], None]:
        """
        A list containing all subject IDs retrieved from ``BIDSLayout`` for timeseries extraction.
        Defined after running ``self.get_bold()``.
        """
        return getattr(self, "_subject_ids", None)

    @property
    def n_cores(self) -> Union[int, None]:
        """
        Number of cores used for multiprocessing with Joblib. Defined after running
        ``self.get_bold()``.
        """
        return getattr(self, "_n_cores", None)

    @property
    def subject_timeseries(self) -> Union[SubjectTimeseries, None]:
        """
        A dictionary mapping subject IDs to their run IDs and their associated timeseries
        (TRs x ROIs) as a NumPy array. Can be deleted using ``del self.subject_timeseries``.
        Defined after running ``self.get_bold()``. This property is also settable (accepts a
        dictionary or pickle file). Returns a reference.
        """
        return getattr(self, "_subject_timeseries", None)

    @subject_timeseries.setter
    def subject_timeseries(self, subject_dict: Union[SubjectTimeseries, str]) -> None:
        subject_dict = io_utils.get_obj(subject_dict)

        self._validate_timeseries(subject_dict)

        self._subject_timeseries = subject_dict

    @subject_timeseries.deleter
    def subject_timeseries(self) -> None:
        del self._subject_timeseries

    @property
    def qc(self) -> Union[dict[str, dict[str, dict[str, Union[float, int]]]], None]:
        """
        A dictionary reporting quality control, which maps subject IDs to their run IDs and
        information related to framewise displacement and dummy scans. Returns a reference.

        ::

            {"subjectID": {"run-ID": {"mean_fd": float, "std_fd": float, ...}}}
        """
        return getattr(self, "_qc", None)

    @staticmethod
    def _validate_timeseries(subject_dict: SubjectTimeseries) -> None:
        """Validates ``subject_timeseries`` structure."""
        error_msg = (
            "A valid pickle file/subject timeseries should contain a nested dictionary where the "
            "first level is the subject id, second level is the run number in the form of 'run-#', "
            "and the final level is the timeseries as a numpy array. "
        )

        error_dict = {
            "Sub": error_msg + "The error occurred at [SUBJECT: {0}]. ",
            "Run": error_msg + "The error occurred at [SUBJECT: {0} | RUN: {1}]. ",
        }

        if not isinstance(subject_dict, dict):
            raise TypeError(error_msg)

        for sub in subject_dict:
            if not isinstance(subject_dict[sub], dict):
                raise TypeError(
                    error_dict["Sub"].format(sub)
                    + "The subject must be a dictionary with second level "
                    "'run-#' keys."
                )

            runs = list(subject_dict[sub])

            if not all("run" in x for x in runs):
                raise TypeError(
                    error_dict["Sub"].format(sub)
                    + "Not all second level keys follow the form of 'run-#'."
                )

            for run in runs:
                if not isinstance(subject_dict[sub][run], np.ndarray):
                    raise TypeError(
                        error_dict["Run"].format(sub, run)
                        + "All 'run-#' keys must contain a numpy array."
                    )

    def _subject_timeseries_size(self) -> str:
        """Computes the byte size of ``self.subject_timeseries``."""
        if not self.subject_timeseries:
            return "0 bytes"

        total_bytes = sum(
            arr.nbytes for subject in self.subject_timeseries.values() for arr in subject.values()
        )
        # Adding size of dictionary
        total_bytes += sys.getsizeof(self.subject_timeseries)

        return f"{total_bytes} bytes"

    def __str__(self) -> str:
        """
        Print Current Object State.

        Provides a formatted summary of the ``TimeseriesExtractor`` configuration when called with
        ``print(self)``. Returns a string containing the following information:

        - Preprocessed BOLD template space
        - Parcellation approach used
        - Signal cleaning parameters applied
        - Task information
        - Number of subjects in ``subject_timeseries``
        - Number of CPU cores used for extraction (multiprocessing)
        - Estimated memory usage estimate for ``subject_timeseries`` (in bytes)

        Returns
        -------
        str
            A formatted string containing information about the object's current state.

        Example
        -------
        >>> from neurocaps.extraction import TimeseriesExtractor
        >>> extractor = TimeseriesExtractor()
        >>> print(extractor)
            Current Object State:
            =====================
            Preprocessed BOLD Template Space                           : "MNI152NLin2009cAsym"
            ...
        """
        # Store some information in variables
        n_subjects = len(self.subject_ids) if self.subject_ids else None
        parcel_approach = list(self.parcel_approach)[0]
        clean_params = self.signal_clean_info
        data_size = self._subject_timeseries_size()

        object_properties = (
            f"Preprocessed BOLD Template Space                           : {self.space}\n"
            f"Parcellation Approach                                      : {parcel_approach}\n"
            f"Signal Cleaning Parameters                                 : {clean_params}\n"
            f"Task Information                                           : {self.task_info}\n"
            f"Number of Subjects                                         : {n_subjects}\n"
            f"CPU Cores Used for Timeseries Extraction (Multiprocessing) : {self.n_cores}\n"
            f"Subject Timeseries Byte Size                               : {data_size}"
        )

        sep = "=" * len(object_properties.rsplit(": ")[0])

        return "Current Object State:\n" + sep + f"\n{object_properties}"



"""Internal module for generating surface plots."""

import os, tempfile
from typing import Any, Union

import nibabel as nib
import matplotlib.pyplot as plt
import surfplot
from neuromaps.datasets import fetch_fslr
from neuromaps.transforms import mni152_to_fslr
from nilearn.plotting.cm import _cmap_d
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from neurocaps.utils import _io as io_utils
from neurocaps.utils._logging import setup_logger
from neurocaps.utils._plot_utils import PlotFuncs

LG = setup_logger(__name__)


def convert_volume_to_surface(
    stat_map: nib.Nifti1Image, fslr_density: str, method: str
) -> tuple[nib.gifti.GiftiImage, nib.gifti.GiftiImage]:
    """
    Converts MNI152 statistical map to GifTI in fsLR surface using neuromaps' ``mni152_to_fslr``.
    """
    # Fix for python 3.12, saving stat map so that it is path instead of a NIfTI
    try:
        gii_lh, gii_rh = mni152_to_fslr(stat_map, method=method, fslr_density=fslr_density)
    except TypeError:
        temp_nifti = create_temp_nifti(stat_map)
        gii_lh, gii_rh = mni152_to_fslr(temp_nifti.name, method=method, fslr_density=fslr_density)
        # Delete
        os.unlink(temp_nifti.name)

    return remove_medial_wall(gii_lh, gii_rh, fslr_density)


def remove_medial_wall(
    gii_lh: nib.gifti.GiftiImage, gii_rh: nib.gifti.GiftiImage, fslr_density: str
) -> tuple[nib.gifti.GiftiImage, nib.gifti.GiftiImage]:
    """Removes medial wall."""
    fslr_atlas = fetch_fslr(density=fslr_density)
    medial_wall_mask = fslr_atlas["medial"]

    gii_lh_mask = nib.load(str(medial_wall_mask[0]))
    gii_lh.darrays[0].data[gii_lh_mask.darrays[0].data == 0] = 0

    gii_rh_mask = nib.load(str(medial_wall_mask[1]))
    gii_rh.darrays[0].data[gii_rh_mask.darrays[0].data == 0] = 0

    return gii_lh, gii_rh


def create_temp_nifti(stat_map: nib.Nifti1Image) -> tempfile._TemporaryFileWrapper:
    """Creates a temporary NIfTI image as a workaround for Python 3.12 issue."""
    # Create temp file
    temp_nii_file = tempfile.NamedTemporaryFile(delete=False, suffix=".nii.gz")
    LG.warning(
        "TypeError raised by neuromaps due to changes in pathlib.py in Python 3.12 "
        "Converting NIfTI image into a temporary nii.gz file (which will be "
        f"automatically deleted afterwards) [TEMP FILE: {temp_nii_file.name}]"
    )

    # Ensure file is closed
    temp_nii_file.close()
    # Save temporary nifti to temp file
    nib.save(stat_map, temp_nii_file.name)

    return temp_nii_file


def generate_surface_plot(
    gii_lh: nib.gifti.GiftiImage,
    gii_rh: nib.gifti.GiftiImage,
    fslr_density: str,
    plot_dict: dict[str, Any],
    group_name: str,
    cap_name: str,
    suffix_title: Union[str, None],
) -> Figure:
    """Creates the surface plot."""
    # Code adapted from example on https://surfplot.readthedocs.io/
    surfaces = fetch_fslr(density=fslr_density)

    if plot_dict["surface"] not in ["inflated", "veryinflated"]:
        LG.warning(
            f"{plot_dict['surface']} is an invalid option for `surface`. Available options "
            "include 'inflated' or 'verinflated'. Defaulting to 'inflated'."
        )
        plot_dict["surface"] = "inflated"

    lh, rh = surfaces[plot_dict["surface"]]
    lh = str(lh) if not isinstance(lh, str) else lh
    rh = str(rh) if not isinstance(rh, str) else rh
    sulc_lh, sulc_rh = surfaces["sulc"]
    sulc_lh = str(sulc_lh) if not isinstance(sulc_lh, str) else sulc_lh
    sulc_rh = str(sulc_rh) if not isinstance(sulc_rh, str) else sulc_rh

    p = surfplot.Plot(
        lh,
        rh,
        size=plot_dict["size"],
        layout=plot_dict["layout"],
        zoom=plot_dict["zoom"],
        views=plot_dict["views"],
        brightness=plot_dict["brightness"],
    )

    # Add base layer
    p.add_layer({"left": sulc_lh, "right": sulc_rh}, cmap="binary_r", cbar=False)
    cmap = _cmap_d[plot_dict["cmap"]] if isinstance(plot_dict["cmap"], str) else plot_dict["cmap"]
    # Add stat map layer
    p.add_layer(
        {"left": gii_lh, "right": gii_rh},
        cmap=cmap,
        alpha=plot_dict["alpha"],
        color_range=plot_dict["color_range"],
        zero_transparent=plot_dict["zero_transparent"],
        as_outline=False,
    )

    if plot_dict["as_outline"] is True:
        p.add_layer(
            {"left": gii_lh, "right": gii_rh},
            cmap="gray",
            cbar=False,
            alpha=plot_dict["outline_alpha"],
            as_outline=True,
        )

    # Color bar
    fig = p.build(
        cbar_kws=plot_dict["cbar_kws"], figsize=plot_dict["figsize"], scale=plot_dict["scale"]
    )
    fig_name = (
        f"{group_name} {cap_name} {suffix_title}" if suffix_title else f"{group_name} {cap_name}"
    )
    fig.axes[0].set_title(fig_name, pad=plot_dict["title_pad"])

    return fig


def save_surface_plot(
    fig: Union[Figure, Axes],
    plot_dict: dict[str, Any],
    output_dir: str,
    plot_output_format: str,
    suffix_filename: Union[str, None],
    stat_map: nib.Nifti1Image,
    save_stat_maps: bool,
    group_name: str,
    cap_name: str,
) -> None:
    """Saves a single surface plot."""
    if not output_dir:
        return None

    filename = io_utils.filename(
        basename=f"{group_name.replace(' ', '_')}_{cap_name}_surface",
        add_name=suffix_filename,
        pos="suffix",
    )
    PlotFuncs.save_fig(fig, plot_dict, output_dir, plot_output_format, filename)

    if save_stat_maps and stat_map:
        filename = filename.split("_surface")[0] + ".nii.gz"
        save_nifti_img(stat_map, output_dir, filename)


def save_nifti_img(stat_map: nib.Nifti1Image, output_dir: str, filename: str) -> None:
    "Save a single NifTI statistical map."
    nib.save(stat_map, os.path.join(output_dir, filename))


def show_surface_plot(fig: Union[Figure, Axes], show_fig: bool) -> None:
    """Visualizes a single surface plot."""
    try:
        plt.show(fig) if show_fig else plt.close(fig)
    except:
        PlotFuncs.show(show_fig)



"""
Internal logging function and class for flushing. Logger configured to respect user logging
configuration prior to package import while defaulting to console logging if no configurations are
provided.
"""

import logging, sys
from logging.handlers import QueueHandler
from multiprocessing.queues import Queue
from typing import Union

# Global variables to determine if a handler is user defined or defined by OS
_PARALLEL_MODULE = "neurocaps.extraction._internals.postprocess"
_USER_ROOT_HANDLER = None
_USER_MODULE_HANDLERS = {}


def setup_logger(
    name: str,
    top_level: bool = True,
    parallel_log_config: Union[dict[str, Union[Queue, int]], None] = None,
):
    """
    Generates module specific loggers, defaults to outputting logs at the informational level to
    the standard output. The ``top_level`` parameter is used to determine if logging configurations
    are present at the time of package import to prevent logging's from sending logs to standard
    error by default.

    Notes
    -----
    **Parallel Processing:** When parallel processing using joblib's loky backend with certain
    modules such as "neurocaps._utils.extraction.extract_timeseries", the logger needs to be
    passed to each child process to ensure to ensure logging configurations are respected. Hence in
    the "extract_timeseries", this function is also called inside ``_extract_timeseries`` function.

    To ensure that there is no issue with race conditions when intending to log to files,
    ``parallel_log_config`` is offered to allow for controlled queue-based logging with
    ``QueueHandler``. This requires a centralized queue with a manager object that is able to share
    reference objects of the centralized queue with each child process. Then ``QueueListener`` can
    be set up in the main process to listen to the centralized queue object and write messages to
    the specified file.
    """
    global _USER_ROOT_HANDLER, _USER_MODULE_HANDLERS

    logger = logging.getLogger(name)

    if top_level is True:
        _USER_ROOT_HANDLER = logging.getLogger().hasHandlers()
        _USER_MODULE_HANDLERS[logger.name] = logger.handlers

    # Special case for parallel config to pass a user-defined logger specifically for parallel
    # processing.
    if logger.name == _PARALLEL_MODULE and not top_level and parallel_log_config:
        logger = setup_queuehandler(logger, parallel_log_config)

    if not logger.level:
        logger.setLevel(logging.INFO)

    # Check if user defined root handler or assigned a specific handler for module
    default_handlers = _USER_ROOT_HANDLER or _USER_MODULE_HANDLERS[logger.name]

    # Propagate to root if no user-defined module or a root handler is detected
    logger.propagate = (
        False if _USER_MODULE_HANDLERS[logger.name] or not _USER_ROOT_HANDLER else True
    )

    # Add or messages will repeat several times due to multiple handlers if same name used
    if not default_handlers and not (
        parallel_log_config or (logger.name == _PARALLEL_MODULE and top_level)
    ):
        logger = add_handler(logger)

    return logger


def setup_queuehandler(logger: logging.Logger, parallel_log_config: dict[str, Union[Queue, int]]):
    # Only QueueHandler will be in handler list
    logger.handlers.clear()
    queue = parallel_log_config.get("queue")
    # Non-strict check
    if not queue:
        ValueError("'queue' is a mandatory key and must contain a queue with a manager object.")
    else:
        logger.addHandler(QueueHandler(queue))

    if "level" in parallel_log_config:
        logger.setLevel(parallel_log_config["level"])

    return logger


def add_handler(logger: logging.Logger, format: Union[str, None] = None):
    """Add and format handler."""
    # Safeguard; ensure a clean state for "extract_timeseries" since it is used in parallel and
    # sequential contexts
    if logger.name == _PARALLEL_MODULE:
        logger.handlers.clear()

    handler = logging.StreamHandler(sys.stdout)

    format = format if format else "%(asctime)s %(name)s [%(levelname)s] %(message)s"
    handler.setFormatter(logging.Formatter(format))
    logger.addHandler(handler)

    return logger



from .parcellations import fetch_preset_parcel_approach, generate_custom_parcel_approach
from .plot_defaults import PlotDefaults

__all__ = ["fetch_preset_parcel_approach", "generate_custom_parcel_approach"]



"""Module containing functions related to fetching the preset parcellation approaches."""

import os
from typing import Literal, Union

import numpy as np

try:
    from nilearn.datasets._utils import fetch_files as _fetch_files
except ImportError:
    from nilearn.datasets.utils import _fetch_files

from ._presets import (
    ATLAS_N_NODES,
    PRESET_ATLAS_NAME,
    PRESET_JSON_NAME,
    PRESET_METADATA,
    OSF_FILE_URLS,
    VALID_PRESETS,
)
from neurocaps.typing import CustomParcelApproach
from neurocaps.utils import _io as io_utils
from neurocaps.utils._helpers import list_to_str
from neurocaps.utils._logging import setup_logger

LG = setup_logger(__name__)


def is_valid_preset(name: str) -> None:
    """Checks if the requested preset name is valid."""
    if not name in VALID_PRESETS:
        # No period
        raise ValueError(
            f"'{name}' is not a supported parcellation. Only the following are supported: "
            f"{list_to_str(VALID_PRESETS)}"
        )


def get_data_dir() -> str:
    """
    Gets the full path for 'neurocaps_data' in the users home directory. If it does not
    exist, then the directory and the subfolders are created.
    """
    data_dir = os.path.expanduser(os.path.join("~", "neurocaps_data"))
    # Create manually instead of using `fetch_files` to notify user that a directory was created
    io_utils.makedir(data_dir)

    return data_dir


def get_preset_subdir_paths(root_dir) -> tuple[str, str]:
    """Returns the names of the subdirs."""
    return os.path.join(root_dir, "presets", "jsons"), os.path.join(root_dir, "presets", "niftis")


def check_n_nodes(name: str, n_nodes: Union[int, None]) -> Union[int, None]:
    """Checks if the number of nodes for the parcellation is valid."""
    has_node_info = name in ATLAS_N_NODES
    if not has_node_info:
        if n_nodes:
            LG.warning(
                f"`n_nodes` parameter is not valid for the {name} parcellation and will be ignored."
            )
        return None

    if not n_nodes:
        n_nodes = ATLAS_N_NODES[name]["default_n"]
        LG.warning(f"Defaulting to {n_nodes} nodes for the {name} parcellation.")
    elif n_nodes not in ATLAS_N_NODES[name]["valid_n"]:
        if not isinstance(n_nodes, int):
            raise ValueError("`n_nodes` must be an integer.")
        else:
            raise ValueError(
                f"{n_nodes} is not valid for the {name} parcellation. "
                f"Only the following are supported: {list_to_str(list(ATLAS_N_NODES.get(name)))}."
            )

    return n_nodes


def get_preset_filenames(name: str, n_nodes: int = None) -> tuple[str, str]:
    """Retrieve the preset filenames."""
    json_filename, nifti_filename = PRESET_JSON_NAME[name], PRESET_ATLAS_NAME[name]
    if name in ATLAS_N_NODES and n_nodes:
        json_filename = json_filename.format(n_nodes)
        nifti_filename = nifti_filename.format(n_nodes)

    return json_filename, nifti_filename


def get_osf_file_url(filename: str) -> str:
    """Retrieves the url for a specific file."""
    base_filename = os.path.basename(filename)

    return r"https://osf.io/{}/download".format(OSF_FILE_URLS[base_filename])


def fetch_custom_parcel_approach(
    name: str, n_nodes: Union[int, None], overwrite: bool = False
) -> dict[Literal["Custom"], CustomParcelApproach]:
    """Fetches the "Custom" parcellation approach. Current valid names are "HCPex" and "4S"."""
    is_valid_preset(name)
    n_nodes = check_n_nodes(name, n_nodes)

    data_dir = get_data_dir()
    json_path, nifti_path = get_preset_subdir_paths(data_dir)

    json_filename, nifti_filename = get_preset_filenames(name, n_nodes)
    filenames = [os.path.join(json_path, json_filename), os.path.join(nifti_path, nifti_filename)]
    fetch_files_from_osf(filenames, overwrite=overwrite)

    # Load in json
    custom_parcel_approach = io_utils.open_json_file(os.path.join(json_path, json_filename))
    # Add path to parcellation file and metadata
    custom_parcel_approach["Custom"]["maps"] = os.path.join(nifti_path, nifti_filename)
    custom_parcel_approach["Custom"]["metadata"] = PRESET_METADATA[name]
    if n_nodes:
        custom_parcel_approach["Custom"]["metadata"]["n_nodes"] = n_nodes

    return custom_parcel_approach


def fetch_files_from_osf(
    filenames: Union[list[str], None],
    overwrite: bool = False,
    resume: bool = True,
    verbose: Union[bool, int] = 1,
) -> None:
    """
    Fetches the files from OSF storage.
    """
    bool_list = [
        io_utils.check_file_exist(filename, raise_errors=False, return_flag=True)
        for filename in filenames
    ]
    if not overwrite:
        filenames = np.array(filenames)
        filenames = filenames[~np.array(bool_list)]
        filenames = filenames.tolist()

    base_filenames = [os.path.basename(filename) for filename in filenames]
    if base_filenames:
        # No period
        LG.info(f"Downloading the following files from OSF: {list_to_str(base_filenames)}")

    opts = {"overwrite": overwrite}
    data_dirs = [os.path.dirname(filename) for filename in filenames]
    download_contents = zip(data_dirs, base_filenames)
    for data_dir, base_filename in download_contents:
        opts["move"] = base_filename
        # (file_path, url, opts)
        files = [(base_filename, get_osf_file_url(base_filename), opts)]
        # Will create data_dir path
        _ = _fetch_files(
            data_dir=data_dir, files=files, resume=resume, verbose=verbose, session=None
        )



"""Module containing helper functions related to visualizing BOLD data."""

import os
from typing import Any, Union

import numpy as np
import matplotlib.pyplot as plt
from numpy.typing import NDArray
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from neurocaps.typing import ParcelApproach
from neurocaps.utils import _io as io_utils
from neurocaps.utils._parcellation_validation import (
    check_parcel_approach,
    extract_custom_region_indices,
    get_parc_name,
)
from neurocaps.utils._plot_utils import PlotFuncs


def get_roi_indices(
    parcel_approach: ParcelApproach, roi_indx: Union[int, str, list[str], list[int]]
) -> NDArray:
    """Gets the indices for a specified node or nodes from ``parcel_approach``."""
    parc_name = get_parc_name(parcel_approach)

    if isinstance(roi_indx, int):
        plot_indxs = roi_indx
    elif isinstance(roi_indx, str):
        # Check if parcellation_approach is custom
        if "Custom" in parcel_approach and "nodes" not in parcel_approach["Custom"]:
            check_parcel_approach(parcel_approach=parcel_approach, call="visualize_bold")

        plot_indxs = list(parcel_approach[parc_name]["nodes"]).index(roi_indx)
    else:
        if all(isinstance(indx, int) for indx in roi_indx):
            plot_indxs = np.array(roi_indx)
        elif all(isinstance(indx, str) for indx in roi_indx):
            # Check if parcellation_approach is custom
            if "Custom" in parcel_approach and "nodes" not in parcel_approach["Custom"]:
                check_parcel_approach(parcel_approach=parcel_approach, call="visualize_bold")

            plot_indxs = np.array(
                [list(parcel_approach[parc_name]["nodes"]).index(index) for index in roi_indx]
            )
        else:
            raise ValueError("All elements in `roi_indx` need to be all strings or all integers.")

    return plot_indxs


def get_region_indices(parcel_approach: ParcelApproach, region: str) -> NDArray:
    """Gets the indices for a specified region from ``parcel_approach``."""
    parc_name = get_parc_name(parcel_approach)
    if "Custom" in parcel_approach:
        if "regions" not in parcel_approach["Custom"]:
            check_parcel_approach(parcel_approach=parcel_approach, call="visualize_bold")
        else:
            plot_indxs = np.array(extract_custom_region_indices(parcel_approach, region))
    else:
        plot_indxs = np.array(
            [
                index
                for index, label in enumerate(parcel_approach[parc_name]["nodes"])
                if region in label
            ]
        )

    return plot_indxs


def get_plot_indxs(
    parcel_approach: ParcelApproach,
    roi_indx: Union[int, str, list[str], list[int]] = None,
    region: str = None,
):
    """Retrieve the indices from the subject's timeseries data to plot."""
    if roi_indx is not None:
        plot_indxs = get_roi_indices(parcel_approach, roi_indx)
    else:
        plot_indxs = get_region_indices(parcel_approach, region)

    return plot_indxs


def create_bold_figure(
    timeseries: NDArray,
    parcel_approach: ParcelApproach,
    figsize: tuple[int, int],
    plot_indxs: NDArray,
    roi_indx: Union[int, str, list[str], list[int]] = None,
    region: str = None,
):
    """Generate the BOLD figure."""
    parc_name = get_parc_name(parcel_approach)
    plt.figure(figsize=figsize)

    if roi_indx or roi_indx == 0:
        plt.plot(range(1, timeseries.shape[0] + 1), timeseries[:, plot_indxs])

        if isinstance(roi_indx, (int, str)) or (isinstance(roi_indx, list) and len(roi_indx) == 1):
            if isinstance(roi_indx, int):
                roi_title = parcel_approach[parc_name]["nodes"][roi_indx]
            elif isinstance(roi_indx, str):
                roi_title = roi_indx
            else:
                roi_title = roi_indx[0]
            plt.title(roi_title)
    else:
        plt.plot(range(1, timeseries.shape[0] + 1), np.mean(timeseries[:, plot_indxs], axis=1))
        plt.title(region)

    plt.xlabel("TR")

    return plt.gcf()


def save_bold_figure(
    fig: Union[Figure, Axes],
    plot_dict: dict[str, Any],
    subj_id: str,
    run_name: str,
    output_dir: str,
    plot_output_format: str,
    filename: str,
):
    """Saves the BOLD figure."""
    if output_dir:
        io_utils.makedir(output_dir)

        if filename:
            save_filename = f"{os.path.splitext(filename.rstrip())[0].rstrip()}"
        else:
            save_filename = f"subject-{subj_id}_{run_name}_timeseries"

        PlotFuncs.save_fig(fig, plot_dict, output_dir, plot_output_format, save_filename)
